import click
import sys

from configurator.hosts import commands as host_commands
from configurator.tasks import commands as task_commands


@click.group()
# @click.version_option(package_name="atkinsonm-configurator")
@click.option(
    "--debug", type=click.BOOL, default=False, is_flag=True, help="Enable debug logging"
)
@click.pass_context
def cli(ctx: click.Context, debug: bool):
    # ensure that ctx.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    ctx.ensure_object(dict)

    ctx.obj["DEBUG"] = debug


@cli.command()
def hello():
    print("Hello from Configurator!")


cli.add_command(host_commands.host)
cli.add_command(task_commands.task)

if __name__ == '__main__':
    args = sys.argv
    cli()



from click.testing import CliRunner
from configurator.cli import cli


def test_hello():
    runner = CliRunner()
    result = runner.invoke(cli, ["hello"])
    assert result.exit_code == 0
    assert result.output == "Hello from Configurator!\n"



from dataclasses import dataclass
from enum import auto, Enum
from typing import Optional
from dataclasses_json import dataclass_json, Undefined


class TaskType(Enum):
    file = auto()
    package = auto()
    file_remove = auto()


@dataclass_json(undefined=Undefined.EXCLUDE)
@dataclass
class Task:
    name: str
    type: TaskType

# This doesn't work well with dataclasses in 3.8, but in 3.10 the kw_only argument may help
# @dataclass_json(undefined=Undefined.EXCLUDE)
# @dataclass(kw_only=True)
# class RestartServiceTask(Task):
#     restart_service: Optional[str] = None


@dataclass
class FileTask(Task):
    """Creates a file with the specified options. DESTRUCTIVE to existing files"""
    path: str
    contents: str
    owner: str
    group: str
    mode: int  # TODO: verify this is a valid permission
    restart_service: Optional[str] = None


@dataclass
class RemoveFileTask(Task):
    """Deletes the file at the specified path."""
    path: str
    restart_service: Optional[str] = None


@dataclass
class PackageTask(Task):
    """Installing and removing packages (Debian support only)"""  # TODO: support other package managers
    package_name: str
    install: bool
    restart_service: Optional[str] = None


task_mapping = {
    "file": FileTask,
    "package": PackageTask,
    "file_remove": RemoveFileTask
}



from pathlib import Path
from typing import List
import click
import os
import yaml

from dataclasses import asdict

from configurator.constants import DEFAULT_HOSTS_FILE_PATH
from configurator.hosts.host import Host


@click.group()
def host():
    """Host configuration activities"""


@host.command()
@click.argument("host-ip", type=click.STRING)
@click.argument("root-password", type=click.STRING)
@click.option(
    "-f",
    "--file",
    "file_path",
    type=Path,
    default=os.path.expandvars(DEFAULT_HOSTS_FILE_PATH),
    show_default=True,
    help="Path to the file which defines hosts",
)
@click.pass_context
def add(
    ctx: click.Context,
    host_ip: str,
    root_password: str,
    file_path: Path,
):
    print(f"Adding host with IP {host_ip} and root password {root_password}")
    print(f"Using file path {file_path}")

    new_host = Host(host_ip, root_password)

    with open(file_path, "r") as hosts_file:
        hosts_dict = yaml.safe_load(hosts_file)
        if hosts_dict and hosts_dict["hosts"]:
            hosts_dict["hosts"].append(asdict(new_host))
        else:
            hosts_dict = {"hosts": [asdict(new_host)]}

    with open(file_path, "w") as hosts_file:
        yaml.safe_dump(hosts_dict, hosts_file)
        # TODO: check for duplicates


@host.command()
@click.argument("host-ip", type=click.STRING)
@click.option(
    "-f",
    "--file",
    "file_path",
    type=Path,
    default=os.path.expandvars(DEFAULT_HOSTS_FILE_PATH),
    show_default=True,
    help="Path to the file which defines hosts",
)
@click.pass_context
def delete(
    ctx: click.Context,
    host_ip: str,
    file_path: Path,
):
    print(f"Deleting host with IP {host_ip}")
    print(f"Using file path {file_path}")

    with open(file_path, "r") as hosts_file:
        hosts_dict = yaml.safe_load(hosts_file)
        if not hosts_dict or not hosts_dict["hosts"]:
            print("Hosts file is empty. Nothing to delete")
        # hosts_dict["hosts"].remove(host_ip)
        new_hosts = []
        for host in hosts_dict["hosts"]:
            if Host(**host).ip_address != host_ip:
                new_hosts.append(host)

    with open(file_path, "w") as hosts_file:
        yaml.safe_dump({"hosts": new_hosts}, hosts_file)


@host.command(name="list")
@click.option(
    "-f",
    "--file",
    "file_path",
    type=Path,
    default=os.path.expandvars(DEFAULT_HOSTS_FILE_PATH),
    show_default=True,
    help="Path to the file which defines hosts",
)
@click.pass_context
def list_command(ctx: click.Context, file_path: Path):
    if not file_path.is_file():
        print("No hosts file found")
        return
    with open(file_path, "r") as hosts_file:
        hosts_dict = yaml.safe_load(hosts_file)
        if not hosts_dict or not hosts_dict["hosts"]:
            print("No hosts found")
            return

        host_list: List[Host] = []

        for host in hosts_dict["hosts"]:
            host_list.append(Host(**host))

        print("-----------------------------------")
        print(f"Hosts configured in {file_path}:")
        print(*host_list, sep="\n")  # TODO: maybe don't dump passwords onto STDOUT
        return host_list



from gettext import install
from paramiko import SSHClient

from configurator.tasks.task import FileTask, PackageTask, RemoveFileTask


def create_file(task: FileTask, client: SSHClient):
    # TODO: add some safety
    print(f"Creating file {task.path}...")
    client.exec_command(f"echo '{task.contents}' > {task.path}")
    client.exec_command(f"chown {task.owner}:{task.group} {task.path}")
    client.exec_command(f"chmod {task.mode} {task.path}")
    if task.restart_service:
        client.exec_command(f"service {task.restart_service} restart")


def delete_file(task: RemoveFileTask, client: SSHClient):
    # TODO: add some safety
    print(f"Deleting file {task.path}...")
    client.exec_command(f"rm {task.path}")
    if task.restart_service:
        client.exec_command(f"service {task.restart_service} restart")


def manage_package(task: PackageTask, client: SSHClient):
    if task.install:
        install_package(task, client)
    else:
        uninstall_package(task, client)
    if task.restart_service:
        client.exec_command(f"service {task.restart_service} restart")


def install_package(task: PackageTask, client: SSHClient):
    print(f"Installing {task.package_name}...")
    stdin, stdout, stderr = client.exec_command(f'apt install {task.package_name} -y')
    # print(stdout)
    # print(stderr)
    stdout.channel.set_combine_stderr(True)
    output = stdout.readlines()  # need this so we can be sure that the operation completes and dpkg lock releases
    # print (output)


def uninstall_package(task: PackageTask, client: SSHClient):
    print(f"Removing {task.package_name}...")
    stdin, stdout, stderr = client.exec_command(f'apt remove {task.package_name} -y')
    # print(stdout)
    # print(stderr)
    stdout.channel.set_combine_stderr(True)
    output = stdout.readlines()  # need this so we can be sure that the operation completes and dpkg lock releases
    # print (output)
    stdin, stdout, stderr = client.exec_command(f'apt autoremove -y')  # autoremove no longer used packages
    # print(stdout)
    # print(stderr)
    stdout.channel.set_combine_stderr(True)
    output = stdout.readlines()  # need this so we can be sure that the operation completes and dpkg lock releases
    # print (output)



# setup.py

from setuptools import setup, find_packages
from io import open
from os import path
import pathlib
# The directory containing this file
HERE = pathlib.Path(__file__).parent
# The text of the README file
README = (HERE / "README.md").read_text()
# automatically captured required modules for install_requires in requirements.txt and as well as configure dependency links
with open(path.join(HERE, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')
install_requires = [x.strip() for x in all_reqs if ('git+' not in x) and (
    not x.startswith('#')) and (not x.startswith('-'))]
dependency_links = [x.strip().replace('git+', '') for x in all_reqs \
                    if 'git+' not in x]

setup(
    name="atkinsonm-configurator",
    description="A simple configuration management tool designed for performing administrative actions on Linux servers",
    version="0.0.4",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires = install_requires,
    entry_points={
        "console_scripts": [
            "atkinsonm-configurator = configurator.cli:cli"
        ]
    },
    author="Mike Atkinson",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/atkinsonm/slack-ops-challenge",
    dependency_links=dependency_links,
    author_email='m.g.atkinson@outlook.com',
)



from typing import List, Optional
import click
import os
from paramiko import AutoAddPolicy, SSHClient
from pathlib import Path
import yaml

from configurator.constants import (
    DEFAULT_HOSTS_FILE_PATH,
    DEFAULT_TASKS_FILE_PATH
)
from configurator.tasks.task import Task, TaskType, task_mapping
from configurator.hosts.commands import list_command as host_list_command
from configurator.hosts.host import Host
from configurator.tasks.task_steps import create_file, delete_file, manage_package


@click.group()
def task():
    """Task activities"""


@task.command(name="list")  # list is a reserved word in Python
@click.option(
    "-f",
    "--file",
    "file_path",
    type=Path,
    default=os.path.expandvars(DEFAULT_TASKS_FILE_PATH),
    show_default=True,
    help="Path to the file which defines tasks",
)
@click.pass_context
def list_command(ctx: click.Context, file_path: Path):
    if not file_path.is_file():
        print("No tasks file found")
        return
    with open(file_path, "r") as tasks_file:
        tasks_dict = yaml.safe_load(tasks_file)
        if not tasks_dict or not tasks_dict["tasks"]:
            print("No tasks found")
            return

        task_list: List[Task] = []

        for task in tasks_dict["tasks"]:
            task_list.append(convert_task_yaml_to_object(task))

        print("-----------------------------------")
        print(f"Tasks configured in {file_path}:")
        print(*task_list, sep="\n")
        return task_list


@task.command()
@click.option(
    "-h",
    "--hosts-file",
    "hosts_file_path",
    type=Path,
    default=os.path.expandvars(DEFAULT_HOSTS_FILE_PATH),
    show_default=True,
    help="Path to the file which defines hosts",
)
@click.option(
    "-t",
    "--tasks-file",
    "tasks_file_path",
    type=Path,
    default=os.path.expandvars(DEFAULT_TASKS_FILE_PATH),
    show_default=True,
    help="Path to the file which defines tasks",
)
@click.pass_context
def run(ctx: click.Context, hosts_file_path: Path, tasks_file_path: Path):
    tasks: List[Task] = ctx.invoke(list_command, file_path=tasks_file_path)
    hosts: List[Host] = ctx.invoke(host_list_command, file_path=hosts_file_path)

    client = SSHClient()
    client.set_missing_host_key_policy(AutoAddPolicy())  # ignore unknown hosts

    for host in hosts:
        # TODO: support a username other than root
        print("-------------------------------------------")
        print(f"Starting tasks for host {host.ip_address}")
        client.connect(host.ip_address, username="root", password=host.root_password)

        for task in tasks:
            if task.type == TaskType.file.name:
                create_file(task, client)

            elif task.type == TaskType.package.name:
                manage_package(task, client)

            elif task.type == TaskType.file_remove.name:
                delete_file(task, client)

        client.close()
        print(f"Completed tasks for host {host.ip_address}")


def convert_task_yaml_to_object(task) -> Optional[Task]:
    valid_task_types = tuple(task_type.name for task_type in TaskType)

    try:
        task_obj = Task(**task)
        if task_obj.type not in valid_task_types:
            print(f"Task {task_obj.name} has invalid type '{task_obj.type}'. Valid values are {valid_task_types}")
        task_type_class = task_mapping.get(task_obj.type, Task)
        return task_type_class(**task)
    except (KeyError, TypeError):
        print(f"Invalid task definition for {task}")
        return None
